from typing import TYPE_CHECKING, overload

from riskit.types import BatchCostFunction, Backend, RiskMetric, Risk, Costs, ArrayLike
from riskit.expected import ExpectedValue
from riskit.variance import MeanVariance
from riskit.sampler import MonteCarlo
from riskit.backend import infer

if TYPE_CHECKING:
    from riskit.backend.numpy import (
        BatchCostFunction as NumpyBatchCostFunction,
        Risk as NumpyRisk,
    )
    from riskit.backend.jax import (
        BatchCostFunction as JaxBatchCostFunction,
        Risk as JaxRisk,
    )


class risk:
    @overload
    @staticmethod
    def expected_value_of[TrajectoriesT, UncertaintySamplesT](
        function: "NumpyBatchCostFunction[TrajectoriesT, UncertaintySamplesT]",
    ) -> RiskMetric[TrajectoriesT, UncertaintySamplesT, "NumpyRisk"]:
        """Creates an Expected Value risk metric using the NumPy backend."""
        ...

    @overload
    @staticmethod
    def expected_value_of[TrajectoriesT, UncertaintySamplesT](
        function: "JaxBatchCostFunction[TrajectoriesT, UncertaintySamplesT]",
    ) -> RiskMetric[TrajectoriesT, UncertaintySamplesT, "JaxRisk"]:
        """Creates an Expected Value risk metric using the JAX backend."""
        ...

    @overload
    @staticmethod
    def expected_value_of[
        TrajectoriesT,
        UncertaintySamplesT,
        CostsT: Costs,
        RiskT: Risk,
        ArrayT: ArrayLike,
    ](
        function: BatchCostFunction[TrajectoriesT, UncertaintySamplesT, CostsT],
        backend: Backend[CostsT, RiskT, ArrayT],
    ) -> RiskMetric[TrajectoriesT, UncertaintySamplesT, RiskT]:
        """Creates an Expected Value risk metric using the specified backend."""
        ...

    @staticmethod
    def expected_value_of[
        TrajectoriesT,
        UncertaintySamplesT,
        CostsT: Costs,
        RiskT: Risk,
        ArrayT: ArrayLike,
    ](
        function: BatchCostFunction[TrajectoriesT, UncertaintySamplesT, CostsT],
        backend: Backend[CostsT, RiskT, ArrayT] | None = None,
    ) -> RiskMetric[TrajectoriesT, UncertaintySamplesT, RiskT]:
        return ExpectedValue(
            cost=function,
            backend=backend
            or infer.backend_from(function, type=Backend[CostsT, RiskT, ArrayT]),
            sampler=MonteCarlo(),
        )

    @overload
    @staticmethod
    def mean_variance_of[TrajectoriesT, UncertaintySamplesT](
        function: "NumpyBatchCostFunction[TrajectoriesT, UncertaintySamplesT]",
        *,
        gamma: float,
    ) -> RiskMetric[TrajectoriesT, UncertaintySamplesT, "NumpyRisk"]:
        """Creates a Mean-Variance risk metric using the NumPy backend."""
        ...

    @overload
    @staticmethod
    def mean_variance_of[
        TrajectoriesT,
        UncertaintySamplesT,
        CostsT: Costs,
        RiskT: Risk,
        ArrayT: ArrayLike,
    ](
        function: BatchCostFunction[TrajectoriesT, UncertaintySamplesT, CostsT],
        *,
        backend: Backend[CostsT, RiskT, ArrayT],
        gamma: float,
    ) -> RiskMetric[TrajectoriesT, UncertaintySamplesT, RiskT]:
        """Creates a Mean-Variance risk metric using the specified backend."""
        ...

    @staticmethod
    def mean_variance_of[
        TrajectoriesT,
        UncertaintySamplesT,
        CostsT: Costs,
        RiskT: Risk,
        ArrayT: ArrayLike,
    ](
        function: BatchCostFunction[TrajectoriesT, UncertaintySamplesT, CostsT],
        *,
        backend: Backend[CostsT, RiskT, ArrayT] | None = None,
        gamma: float,
    ) -> RiskMetric[TrajectoriesT, UncertaintySamplesT, RiskT]:
        return MeanVariance(
            cost=function,
            backend=backend
            or infer.backend_from(function, type=Backend[CostsT, RiskT, ArrayT]),
            sampler=MonteCarlo(),
            gamma=gamma,
        )



from typing import Protocol, NamedTuple
from types import TracebackType


class ArrayLike(Protocol):
    @property
    def shape(self) -> tuple[int, ...]:
        """Shape of the array like object."""
        ...


type Costs = ArrayLike
"""A costs array of shape (trajectories, uncertainty samples)."""

type Risk = ArrayLike
"""A risk array of shape (trajectories,)."""


class BatchCostFunction[TrajectoriesT, UncertaintySamplesT, CostsT: Costs](Protocol):
    def __call__(
        self, *, trajectories: TrajectoriesT, uncertainties: UncertaintySamplesT
    ) -> CostsT:
        """Describes the cost function that should be used for evaluating risk.

        Returns:
            An array like object of shape (M, N) where M is the number of trajectories
            and N is the number of uncertainty samples.
        """
        ...


class TrajectoriesProvider[TrajectoriesT](Protocol):
    async def get(self) -> TrajectoriesT:
        """Provides the batch of trajectories for which the risk metric should be computed."""
        ...

    @property
    def trajectory_count(self) -> int:
        """Number of trajectories provided."""
        ...


class Uncertainties[UncertaintySamplesT](Protocol):
    async def sample(self, count: int) -> UncertaintySamplesT:
        """Returns samples from the distribution(s) of the uncertain variable(s)."""
        ...


class Computation[CostsT: Costs, RiskT: Risk, ArrayT: ArrayLike](Protocol):
    async def array_of(self, costs: CostsT) -> ArrayT:
        """Converts the given costs to the backend's array type."""
        ...

    async def to_risk(self, array: ArrayT) -> RiskT:
        """Converts the given array to the backend's risk type."""
        ...

    async def mean(self, data: ArrayT, *, axis: int) -> ArrayT:
        """Computes the mean of the given data along the specified axis."""
        ...

    async def var(self, data: ArrayT, *, axis: int) -> ArrayT:
        """Computes the variance of the given data along the specified axis."""
        ...

    async def axpby(
        self, *, a: float = 1.0, x: ArrayT, b: float = 1.0, y: ArrayT
    ) -> ArrayT:
        """Computes the operation a * x + b * y."""
        ...


class Backend[CostsT: Costs, RiskT: Risk, ArrayT: ArrayLike](Protocol):
    async def __aenter__(self) -> Computation[CostsT, RiskT, ArrayT]:
        """Enters the backend context for computations."""
        ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exits the backend context."""
        ...


class SamplingResult[UncertaintySamplesT](NamedTuple):
    samples: UncertaintySamplesT
    sample_count: int


class Sampler[UncertaintySamplesT](Protocol):
    async def sample_from(
        self, uncertainties: Uncertainties[UncertaintySamplesT]
    ) -> SamplingResult[UncertaintySamplesT]:
        """Samples from the given uncertainties."""
        ...


class RiskMetric[TrajectoriesT, UncertaintySamplesT, RiskT: Risk](Protocol):
    async def compute(
        self,
        *,
        trajectories: TrajectoriesProvider[TrajectoriesT],
        uncertainties: Uncertainties[UncertaintySamplesT],
    ) -> RiskT:
        """Computes the risk metric for the given trajectories and uncertainties.

        Returns:
            An array like object of shape (trajectories,) representing the risk
            values for each trajectory.
        """
        ...

    def sampled_with(
        self, sampler: Sampler[UncertaintySamplesT]
    ) -> "RiskMetric[TrajectoriesT, UncertaintySamplesT, RiskT]":
        """Returns a new risk metric that uses the given sampler to sample from uncertainties."""
        ...



from typing import cast
from dataclasses import dataclass

from riskit.backend.jax.backend import Uncertainty, UncertaintySamples

import jax.numpy as jnp
import jax.random as jrandom
from jax import Array

# TODO: Extract seeds to be configurable.


@dataclass(frozen=True)
class Gaussian(Uncertainty):
    mean: Array
    covariance: Array

    def __post_init__(self) -> None:
        V, D = self.mean.shape

        assert self.covariance.shape == (V, D, D), (
            f"Covariance shape {self.covariance.shape} does not match mean shape {self.mean.shape}. "
            f"Expected covariance shape to be {(V, D, D)}, got {self.covariance.shape}."
        )

    async def sample(self, count: int) -> UncertaintySamples:
        V, D = self.mean.shape

        key = jrandom.PRNGKey(0)

        samples_list = []
        for v in range(V):
            key, subkey = jrandom.split(key)
            samples_v = jrandom.multivariate_normal(
                subkey, self.mean[v], self.covariance[v], shape=(count,)
            )
            samples_list.append(samples_v)

        samples = jnp.stack(samples_list, axis=0).transpose(0, 2, 1)

        assert samples.shape == (V, D, count), (
            f"Samples shape {samples.shape} does not match expected shape {(V, D, count)}."
        )

        return samples

    @property
    def shape(self) -> tuple[int, int]:
        return cast(tuple[int, int], self.mean.shape)


@dataclass(frozen=True)
class Uniform(Uncertainty):
    lower: Array
    upper: Array

    def __post_init__(self) -> None:
        lower_shape = self.lower.shape

        assert self.upper.shape == lower_shape, (
            f"Upper bound shape {self.upper.shape} does not match lower bound shape {self.lower.shape}. "
            f"Expected upper bound shape to be {lower_shape}, got {self.upper.shape}."
        )
        assert jnp.all(self.upper >= self.lower), (
            "All upper bounds must be greater than lower bounds."
        )

    async def sample(self, count: int) -> UncertaintySamples:
        key = jrandom.PRNGKey(0)

        samples = self._roll_axes_left(
            jrandom.uniform(
                key,
                shape=(count, *(shape := self.lower.shape)),
                minval=self.lower,
                maxval=self.upper,
            )
        )

        assert samples.shape == (*shape, count), (
            f"Samples shape {samples.shape} does not match expected shape {(*shape, count)}."
        )

        return samples

    @property
    def shape(self) -> tuple[int, ...]:
        return self.lower.shape

    def _roll_axes_left(self, array: Array) -> Array:
        rest = array.shape[1:]
        return array.transpose(*range(1, len(rest) + 1), 0)


class distribution:
    @staticmethod
    def gaussian(*, mean: Array, covariance: Array) -> Gaussian:
        return Gaussian(mean=mean, covariance=covariance)

    @staticmethod
    def uniform(*, lower: Array, upper: Array) -> Uniform:
        return Uniform(lower=lower, upper=upper)



from dataclasses import dataclass

from riskit.types import (
    Costs,
    Risk,
    ArrayLike,
    BatchCostFunction,
    TrajectoriesProvider,
    Uncertainties,
    Backend,
    Sampler,
)


@dataclass(frozen=True)
class ExpectedValue[
    TrajectoriesT,
    UncertaintySamplesT,
    CostsT: Costs,
    RiskT: Risk,
    ArrayT: ArrayLike,
]:
    cost: BatchCostFunction[TrajectoriesT, UncertaintySamplesT, CostsT]
    backend: Backend[CostsT, RiskT, ArrayT]
    sampler: Sampler[UncertaintySamplesT]

    async def compute(
        self,
        *,
        trajectories: TrajectoriesProvider[TrajectoriesT],
        uncertainties: Uncertainties[UncertaintySamplesT],
    ) -> RiskT:
        M = trajectories.trajectory_count

        samples, N = await self.sampler.sample_from(uncertainties)
        costs = self.cost(trajectories=await trajectories.get(), uncertainties=samples)

        assert costs.shape == (M, N), (
            f"Costs shape {costs.shape} does not match expected shape {(M, N)}."
        )

        async with self.backend as op:
            return await op.to_risk(await op.mean(await op.array_of(costs), axis=1))

    def sampled_with(
        self, sampler: Sampler[UncertaintySamplesT]
    ) -> "ExpectedValue[TrajectoriesT, UncertaintySamplesT, CostsT, RiskT, ArrayT]":
        return ExpectedValue(cost=self.cost, backend=self.backend, sampler=sampler)



from dataclasses import dataclass

from riskit.types import Uncertainties, SamplingResult


@dataclass(frozen=True)
class MonteCarlo[UncertaintySamplesT]:
    sample_count: int = 5000

    async def sample_from(
        self, uncertainties: Uncertainties[UncertaintySamplesT]
    ) -> SamplingResult[UncertaintySamplesT]:
        return SamplingResult(
            samples=await uncertainties.sample(count=self.sample_count),
            sample_count=self.sample_count,
        )



from typing import Protocol

from riskit.types import BatchCostFunction, Backend, Costs
from riskit.annotations import return_annotation_of


class BackendCreator[BackendT: Backend](Protocol):
    def __call__(self) -> BackendT:
        """Creates a backend instance for computations."""
        ...


def backend_from[
    TrajectoriesT,
    UncertaintySamplesT,
    CostsT: Costs,
    BackendT: Backend,
](
    function: BatchCostFunction[TrajectoriesT, UncertaintySamplesT, CostsT],
    *,
    type: type[BackendT],
) -> BackendT:
    return return_annotation_of(
        function, type=tuple[BackendCreator[BackendT]]
    ).metadata[0]()



import inspect
from typing import Callable, cast
from dataclasses import dataclass


@dataclass(frozen=True)
class ExtractedAnnotation[*MetadataT]:
    metadata: tuple[*MetadataT]


def return_annotation_of[*MetadataT](
    function: Callable, *, type: type[tuple[*MetadataT]]
) -> ExtractedAnnotation[*MetadataT]:
    signature = inspect.signature(function)
    annotation = signature.return_annotation

    try:
        metadata = annotation.__value__.__metadata__
        return ExtractedAnnotation(metadata=cast(tuple[*MetadataT], metadata))
    except AttributeError as e:
        # TODO: Test me!
        raise TypeError(
            f"Could not extract annotation metadata from {annotation} of function {function}. "
            "Ensure the return type is properly annotated."
        ) from e



from .backend import (
    Inputs as Inputs,
    States as States,
    InputAndState as InputAndState,
    InputAndStateProvider as InputAndStateProvider,
    Uncertainty as Uncertainty,
    UncertaintySamples as UncertaintySamples,
    Costs as Costs,
    Risk as Risk,
    BatchCostFunction as BatchCostFunction,
    JaxBackend as JaxBackend,
)

from .distribution import distribution as distribution



from typing import cast
from dataclasses import dataclass

from riskit.backend.numpy.backend import Uncertainty, UncertaintySamples

import numpy as np
from numtypes import Array, shape_of, Dims, Shape


# NOTE: We explicitly implement the Uncertainty protocol to help the type checker.
# TODO: Add seeds for reproducibility.


@dataclass(frozen=True)
class Gaussian[V: int, D: int](Uncertainty[Dims[V, D]]):
    mean: Array[Dims[V, D]]
    covariance: Array[Dims[V, D, D]]

    def __post_init__(self) -> None:
        V, D = self.mean.shape

        assert shape_of(self.covariance, matches=(V, D, D)), (
            f"Covariance shape {self.covariance.shape} does not match mean shape {self.mean.shape}. "
            f"Expected covariance shape to be {(V, D, D)}, got {self.covariance.shape}."
        )

    async def sample[N: int](self, count: N) -> UncertaintySamples[Dims[V, D], N]:
        V, D = self.mean.shape

        samples = np.array(
            [
                np.random.multivariate_normal(
                    self.mean[v], self.covariance[v], size=count
                )
                for v in range(V)  # TODO: For loop is not the best.
            ]
        ).transpose(0, 2, 1)

        assert shape_of(samples, matches=(V, D, count), name="samples")

        return samples

    @property
    def shape(self) -> Dims[V, D]:
        return self.mean.shape


@dataclass(frozen=True)
class Uniform[ShapeT: Shape](Uncertainty[ShapeT]):
    lower: Array[ShapeT]
    upper: Array[ShapeT]

    def __post_init__(self) -> None:
        lower_shape = self.lower.shape

        assert shape_of(self.upper, matches=lower_shape), (
            f"Upper bound shape {self.upper.shape} does not match lower bound shape {self.lower.shape}. "
            f"Expected upper bound shape to be {lower_shape}, got {self.upper.shape}."
        )
        assert np.all(self.upper >= self.lower), (
            "All upper bounds must be greater than lower bounds."
        )

    async def sample[N: int](self, count: N) -> UncertaintySamples[ShapeT, N]:
        samples = self._roll_axes_left(
            np.random.uniform(
                low=self.lower,
                high=self.upper,
                size=(count, *(shape := self.lower.shape)),
            )
        )

        assert shape_of(samples, matches=(*shape, count), name="samples")

        return cast(UncertaintySamples[ShapeT, N], samples)

    @property
    def shape(self) -> ShapeT:
        return self.lower.shape

    def _roll_axes_left(self, array: Array) -> Array:
        rest = array.shape[1:]
        return array.transpose(*range(1, len(rest) + 1), 0)


class distribution:
    @staticmethod
    def gaussian[V: int, D: int](
        *, mean: Array[Dims[V, D]], covariance: Array[Dims[V, D, D]]
    ) -> Gaussian[V, D]:
        return Gaussian(mean=mean, covariance=covariance)

    @staticmethod
    def uniform[ShapeT: Shape](
        *, lower: Array[ShapeT], upper: Array[ShapeT]
    ) -> Uniform[ShapeT]:
        return Uniform(lower=lower, upper=upper)



from typing import Protocol, NamedTuple, Self, Annotated
from types import TracebackType

from numtypes import Array, Dim1, Dim2, Dim3, Dims, Shape, shape_of

import numpy as np


class NumpyBackend:
    @staticmethod
    def create() -> "NumpyBackend":
        return NumpyBackend()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        pass

    async def array_of(self, costs: "Costs") -> "Costs":
        return costs

    async def to_risk(self, array: Array) -> "Risk":
        assert shape_of(array, matches=(-1,))

        return array

    async def mean(self, data: Array, *, axis: int) -> Array:
        return np.mean(data, axis=axis)

    async def var(self, data: Array, *, axis: int) -> Array:
        return np.var(data, axis=axis)

    async def axpby(
        self, *, a: float = 1.0, x: Array, b: float = 1.0, y: Array
    ) -> Array:
        return a * x + b * y


type Inputs = Array[Dim3]
"""Batch of control inputs. The dimensions are (time steps, control dimensions, trajectories)."""

type States = Array[Dim3]
"""Batch of states. The dimensions are (time steps, state dimensions, trajectories)."""

type UncertaintySamples[ShapeT: Shape = Shape, N: int = int] = Array[Dims[*ShapeT, N]]
"""Batch of uncertainty samples. The dimensions are (...uncertainty dimensions, uncertainty samples)."""

type Costs = Annotated[Array[Dim2], NumpyBackend.create]
"""Batch of costs. The dimensions are (trajectories, uncertainty samples)."""

type Risk = Array[Dim1]
"""Batch of risk values. The dimensions are (trajectories,)."""


class Uncertainty[ShapeT: Shape](Protocol):
    async def sample[N: int](self, count: N) -> UncertaintySamples[ShapeT, N]:
        """Returns samples from the distribution of the uncertain variable(s)."""
        ...


class BatchCostFunction[TrajectoriesT, UncertaintySamplesT](Protocol):
    def __call__(
        self, *, trajectories: TrajectoriesT, uncertainties: UncertaintySamplesT
    ) -> Costs:
        """Describes the cost function that should be used for evaluating risk."""
        ...


class InputAndState(NamedTuple):
    u: Inputs
    x: States

    async def get(self) -> Self:
        return self

    @property
    def trajectory_count(self) -> int:
        return self.u.shape[2]



from .backend import numpy as numpy, jax as jax, distribution as distribution
from .types import (
    Costs as Costs,
    Risk as Risk,
    Uncertainties as Uncertainties,
    RiskMetric as RiskMetric,
    SamplingResult as SamplingResult,
)
from .risk import risk as risk
from .sampler import MonteCarlo as MonteCarlo
